/** 
 * \file loop_patch.cpp
 *  
 * \brief  Implementation of  the class  LOOPatch, which  represents a
 * Loop subdivision surface patch.
 *
 * \author
 * Marcelo Ferreira Siqueira \n
 * Universidade Federal do Rio Grande do Norte, \n
 * Departamento de Informatica e Matematica Aplicada, \n
 * marcelo at dimap (dot) ufrn (dot) br
 *
 * \version 2.0
 * \date January 2010
 *
 * \attention This program is distributed WITHOUT ANY WARRANTY, and it
 *            may be freely redistributed under the condition that the
 *            copyright notices  are not removed,  and no compensation
 *            is received. Private, research, and institutional use is
 *            free. Distribution of this  code as part of a commercial
 *            system  is permissible ONLY  BY DIRECT  ARRANGEMENT WITH
 *            THE AUTHOR.
 */

#include "loop_patch.h"  // LOOPatch

#include <cassert>       // assert
#include <cmath>         // fabs


/** 
* \defgroup PPSFROMLOOPNameSpace Namespace ppsfromloop.
* @{
*/

/**
* \namespace ppsfromloop
*
* \brief  The  namespace   ppsfromloop  contains  the  definition  and
* implementation   of   all    classes   representing   a   Parametric
* Pseudo-Surface (PPS) that approximates  the geometry of a given Loop
* subdivision surface.
*/


namespace ppsfromloop {

  /**
   * \fn LOOPatch::LOOPatch( EvalStruct* evs , unsigned evd , BCORDER bc , double* pp ) 
   *
   * \brief Creates an instance of this class.
   *
   * \param  evs  Pointer to  a  Loop  subdivision surface  evaluator.
   * \param evd  Degree of the  only extraordinary vertex (if  any) of
   * the base  mesh face  associated with this  face.  If there  is no
   * extraordinary vertex,  the default parameter value  is 6.  
   * \param bc  Order of  the barycentric coordinates  associated with
   * the affine  frame corresponding to the base  mesh face associated
   * with this patch.
   * \param pp Pointer to the array of projected points of this patch.
   */
  LOOPatch::LOOPatch(
		     EvalStruct* evs ,
		     unsigned evd ,
		     BCORDER bc ,
		     double* pp 
		    )
    :
      _evstruct( evs ) ,
      _evdegree( evd ) ,
      _bcorder( bc )
  {

    const unsigned int K = get_extraordinary_vertex_degree() + 6 ;
      
    _proj_pts = new double[ K * 3 ] ;
      
    unsigned i;
    for ( i = 0 ; i < K * 3 ; i++ ) {
      _proj_pts[ i ] = pp[ i ] ;
    }
    
    return ;
  }


  /**
   * \fn LOOPatch::LOOPatch( const LOOPatch& patch ) 
   *
   * \brief Creates an instance of this class from another instance.
   *
   * \param patch An instance of this class.
   */
  LOOPatch::LOOPatch( const LOOPatch& patch ) 
  {
    if ( &patch != this ) {
      _evstruct = patch._evstruct ;
      _evdegree = patch._evdegree ;
      _bcorder = patch._bcorder ;
      
      const unsigned int K = get_extraordinary_vertex_degree() + 6 ;
      
      _proj_pts = new double[ K * 3 ] ;
      
      unsigned i;
      for ( i = 0 ; i < K * 3 ; i++ ) {
	_proj_pts[ i ] = patch._proj_pts[ i ] ;
      }
    }

    return ;
  }


  /**
   * \fn void LOOPatch::point( double u , double v , double& x , double& y , double& z ) const 
   *
   * \brief Computes a point on this patch.
   *
   * \param u First barycentric coordinate of a given parameter point.
   * \param v Second barycentric coordinate of a given parameter point.
   * \param x First Cartesian coordinate of the point on this patch.
   * \param y Second Cartesian coordinate of the point on this patch.
   * \param z Third Cartesian coordinate of the point on this patch.
   *
   */
  void
  LOOPatch::point( 
		  double u , 
		  double v , 
		  double& x , 
		  double& y , 
		  double& z 
		 ) 
    const
  {
    if ( fabs( u ) <= 1e-15 ) {
      u = 0 ;
    }
    else if ( fabs( 1 - u ) <= 1e-15 ) {
      u = 1 ;
    }

    if ( fabs( v ) <= 1e-15 ) {
      v = 0 ;
    }
    else if ( fabs( 1 - v ) <= 1e-15 ) {
      v = 1 ;
    }

    double w = 1 - ( u + v ) ;

    if ( fabs( w ) <= 1e-15 ) {
      w = 0 ;
    }
    else if ( fabs( 1 - w ) <= 1e-15 ) {
      w = 1 ;
    }

    double vv, ww;

    if ( get_barycentric_coordinates_order() == UVW ) {
      vv = v;
      ww = w;
    }
    else if ( get_barycentric_coordinates_order() == VWU ) {
      vv = w;
      ww = u;
    }
    else {
      vv = u;
      ww = v;
    }

    double s[3][3];
    get_surface_evaluator()->eval_surf( 
				       EvalStruct::EVAL_VALUE , 
				       get_projected_points() ,
				       get_extraordinary_vertex_degree() ,
				       vv ,
				       ww ,
				       s 
				      ) ;

    x = s[ 0 ][ 0 ] ;
    y = s[ 0 ][ 1 ] ;
    z = s[ 0 ][ 2 ] ;
    
    return ;
  }

}

/** @} */ //end of group class.
